// Copyright (c) Mito
import React from 'react';
import { isFilterGroup } from './filterTypes';
import { getEmptyFilterData } from './utils';
import { Filter } from './Filter';
import FilterGroup from './FilterGroup';
import { CONDITIONS_WITH_NO_INPUT } from './filterConditions';
import DropdownButton from '../../../elements/DropdownButton';
import Select from '../../../elements/Select';
import Row from '../../../spacing/Row';
import Col from '../../../spacing/Col';
import '../../../../../css/taskpanes/ControlPanel/FilterCard.css';
/*
    We wait 500ms before sending a filter message to make sure
    that as the user is typing key changes, we don't queue up a
    ton of filtering messages.
*/
const FILTER_MESSAGE_DELAY = 500;
const ADD_FILTER_SELECT_TITLE = '+ Add Filter';
/*
    Component that contains all that one needs to filter!
*/
class FilterCard extends React.Component {
    constructor(props) {
        super(props);
        this.state = {
            filters: props.filters,
            operator: props.operator,
            stepID: '',
            filterSelect: ADD_FILTER_SELECT_TITLE,
            timeoutId: undefined
        };
        this.sendFilterUpdateMessage = this.sendFilterUpdateMessage.bind(this);
        this._sendFilterUpdateMessage = this._sendFilterUpdateMessage.bind(this);
        this.addFilter = this.addFilter.bind(this);
        this.addFilterGroup = this.addFilterGroup.bind(this);
        this.addFilterToGroup = this.addFilterToGroup.bind(this);
        this.deleteFilter = this.deleteFilter.bind(this);
        this.deleteFilterFromGroup = this.deleteFilterFromGroup.bind(this);
        this.setFilter = this.setFilter.bind(this);
        this.setFilterInGroup = this.setFilterInGroup.bind(this);
        this.setOperator = this.setOperator.bind(this);
        this.setOperatorInGroup = this.setOperatorInGroup.bind(this);
    }
    /*
        We only send a filter message out after .5 seconds, and delete any
        message that has not been sent if a new one comes in. This makes sure we don't
        have a huge backlog of filter message waiting to be processed
    */
    async sendFilterUpdateMessage() {
        if (this.state.timeoutId) {
            global.clearTimeout(this.state.timeoutId);
        }
        const newTimeoutId = global.setTimeout(this._sendFilterUpdateMessage, FILTER_MESSAGE_DELAY);
        this.setState({ timeoutId: newTimeoutId });
    }
    /*
        NOTE: only call this through the sendFilterUpdateMessage function, to make sure
        buffering messages works properly.

        Before sending the displayed filters, we parse all the number filters from strings
        to numbers, and then filter out all of the invalid filters (as to not cause errors)
    */
    async _sendFilterUpdateMessage() {
        /*
            A filter is invalid if:
            1. It should have an input, and it does not
            2. It is a number filter with a string input, or with a non-valid number input
        */
        const isValidFilter = (filter) => {
            if (CONDITIONS_WITH_NO_INPUT.includes(filter.condition)) {
                return true;
            }
            if (filter.type === "number_series" /* NUMBER_SERIES */) {
                return typeof filter.value !== 'string' && !isNaN(filter.value);
            }
            if (filter.type === "string_series" /* STRING_SERIES */ || filter.type === "datetime_series" /* DATETIME_SERIES */) {
                return filter.value !== '';
            }
            return true;
        };
        /*
            The frontend stores number filters as strings, and so we parse them to
            numbers before sending them to the backend
        */
        const parseFilter = (filter) => {
            if (filter.type === "number_series" /* NUMBER_SERIES */ && typeof filter.value === 'string') {
                return {
                    type: filter.type,
                    condition: filter.condition,
                    value: parseFloat(filter.value)
                };
            }
            return filter;
        };
        // To handle decimals, we allow decimals to be submitted, and then just
        // parse them before they are sent to the back-end
        const parsedFilters = this.state.filters.map((filterOrGroup) => {
            if (isFilterGroup(filterOrGroup)) {
                return {
                    filters: filterOrGroup.filters.map((filter) => {
                        return parseFilter(filter);
                    }),
                    operator: filterOrGroup.operator
                };
            }
            else {
                return parseFilter(filterOrGroup);
            }
        });
        const filtersToApply = parsedFilters.map((filterOrGroup) => {
            // Filter out these incomplete filters from the group
            if (isFilterGroup(filterOrGroup)) {
                return {
                    filters: filterOrGroup.filters.filter((filter) => {
                        return isValidFilter(filter);
                    }),
                    operator: filterOrGroup.operator
                };
            }
            else {
                return filterOrGroup;
            }
        }).filter((filterOrGroup) => {
            // Filter out the groups if they have no valid filters in them
            if (isFilterGroup(filterOrGroup)) {
                return filterOrGroup.filters.length > 0;
            }
            // And then we filter the non group filters to be non-empty
            return isValidFilter(filterOrGroup);
        });
        const stepID = await this.props.mitoAPI.sendFilterMessage(this.props.selectedSheetIndex, this.props.columnHeader, filtersToApply, this.state.operator, this.state.stepID);
        this.setState({ stepID: stepID });
    }
    /*
        Adds a new, blank filter to the end of the filters list
    */
    addFilter() {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            newFilters.push(getEmptyFilterData(this.props.columnMitoType));
            return {
                filters: newFilters
            };
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Creates a new filter group (at the bottom) with a single empty
        filter
    */
    addFilterGroup() {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            newFilters.push({
                filters: [
                    getEmptyFilterData(this.props.columnMitoType)
                ],
                operator: 'And'
            });
            return {
                filters: newFilters
            };
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Adds a blank new filter to the end of a specific group
    */
    addFilterToGroup(groupIndex) {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            const filterGroup = newFilters[groupIndex];
            if (isFilterGroup(filterGroup)) {
                // If we do have a filter group at that groupIndex, then we add a new filter to it
                filterGroup.filters.push(getEmptyFilterData(this.props.columnMitoType));
                return {
                    filters: newFilters
                };
            }
            else {
                // We make no changes if this was not a filter group, which should never occur
                return null;
            }
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Deletes a filter that is at the given index in the main
        filter list.
    */
    deleteFilter(filterIndex) {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            newFilters.splice(filterIndex, 1);
            return {
                filters: newFilters
            };
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Deletes a filter that is at a given index in filter list
        of a specific filter group
    */
    deleteFilterFromGroup(groupIndex, filterIndex) {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            const filterGroup = newFilters[groupIndex];
            if (isFilterGroup(filterGroup)) {
                // If we do have a filter group at that groupIndex, then we delete the filter
                // at the passed filterIndex
                filterGroup.filters.splice(filterIndex, 1);
                // If there are no filters left in this group, then we remove the entire group
                if (filterGroup.filters.length === 0) {
                    newFilters.splice(groupIndex, 1);
                }
                return {
                    filters: newFilters
                };
            }
            else {
                // We make no changes if this was not a filter group, which should never occur
                return null;
            }
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Sets a filter at the given index to the new filter value
    */
    setFilter(filterIndex, filter) {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            newFilters[filterIndex] = filter;
            return {
                filters: newFilters
            };
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Sets a filter at the given filterIndex in the specific group at the given
        groupIndex to the new filter value
    */
    setFilterInGroup(groupIndex, filterIndex, filter) {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            const filterGroup = newFilters[groupIndex];
            if (isFilterGroup(filterGroup)) {
                filterGroup.filters[filterIndex] = filter;
                return {
                    filters: newFilters
                };
            }
            else {
                // We make no changes if this was not a filter group, which should never occur
                return null;
            }
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Sets the operator that combines the highest level filters and
        filter groups
    */
    setOperator(operator) {
        this.setState({ operator: operator }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    /*
        Sets the operator that combines a specific filter group
    */
    setOperatorInGroup(groupIndex, operator) {
        this.setState(prevState => {
            const newFilters = [...prevState.filters];
            const filterGroup = newFilters[groupIndex];
            if (isFilterGroup(filterGroup)) {
                filterGroup.operator = operator;
                return {
                    filters: newFilters
                };
            }
            else {
                // We make no changes if this was not a filter group, which should never occur
                return null;
            }
        }, () => {
            void this.sendFilterUpdateMessage();
        });
    }
    render() {
        return (React.createElement("div", null,
            React.createElement("div", { className: 'text-header-3 mt-15px' },
                React.createElement("p", null, " Filter ")),
            this.state.filters.map((filterOrGroup, index) => {
                if (isFilterGroup(filterOrGroup)) {
                    return (
                    /*
                        If the FilterGroup is the first Filter or FilterGroup
                        in the FilterCard, add a 'Where'
                    */
                    React.createElement(Row, { justify: 'space-between', align: 'top' },
                        React.createElement(Col, { span: 4 },
                            index === 0 &&
                                React.createElement("p", { className: 'text-body-1' }, "Where"),
                            index !== 0 &&
                                React.createElement(Select, { value: this.state.operator, options: ['And', 'Or'], onChange: (newOperator) => this.setOperator(newOperator) })),
                        React.createElement(Col, { span: 19 },
                            React.createElement(FilterGroup, { key: index, mainOperator: this.state.operator, filters: filterOrGroup.filters, groupOperator: filterOrGroup.operator, setFilter: (filterIndex, newFilter) => {
                                    this.setFilterInGroup(index, filterIndex, newFilter);
                                }, setOperator: (newOperator) => {
                                    this.setOperatorInGroup(index, newOperator);
                                }, deleteFilter: (filterIndex) => {
                                    this.deleteFilterFromGroup(index, filterIndex);
                                }, addFilter: () => this.addFilterToGroup(index) }))));
                }
                else {
                    return (React.createElement(Filter, { first: index === 0, key: index, filter: filterOrGroup, operator: this.state.operator, displayOperator: true, setFilter: (newFilter) => {
                            this.setFilter(index, newFilter);
                        }, setOperator: this.setOperator, deleteFilter: () => { this.deleteFilter(index); } }));
                }
            }),
            React.createElement(DropdownButton, { options: {
                    'Add a Filter': this.addFilter,
                    'Add a Group of Filters': this.addFilterGroup
                }, width: 'medium' }, ADD_FILTER_SELECT_TITLE)));
    }
}
export default FilterCard;
//# sourceMappingURL=FilterCard.js.map