// Copyright (c) Mito
import React, { Fragment } from 'react';
import DefaultTaskpane from '../DefaultTaskpane';
import MultiToggleBox from '../../elements/MultiToggleBox';
import { getSuggestedColumnHeaderKeys } from './mergeUtils';
import MergeSheetAndKeySelection from './MergeSheetAndKeySelection';
import Row from '../../spacing/Row';
import Col from '../../spacing/Col';
import Select from '../../elements/Select';
// Enum to allow you to refer to the first or second sheet by name, for clarity
export var MergeSheet;
(function (MergeSheet) {
    MergeSheet[MergeSheet["First"] = 0] = "First";
    MergeSheet[MergeSheet["Second"] = 1] = "Second";
})(MergeSheet || (MergeSheet = {}));
/*
    Each entry of this enum is a merge type that the user can choose.
    In all cases, except lookup, these values are passed directly to
    the pandas merge function.
*/
export var MergeType;
(function (MergeType) {
    MergeType["LOOKUP"] = "lookup";
    MergeType["LEFT"] = "left";
    MergeType["RIGHT"] = "right";
    MergeType["INNER"] = "inner";
    MergeType["OUTER"] = "outer";
})(MergeType || (MergeType = {}));
class MergeTaskpane extends React.Component {
    constructor(props) {
        super(props);
        // We default the merge modal to select the selected sheet index as the first sheet
        const sheetOneIndex = props.selectedSheetIndex;
        // The second sheet is either: the sheet to the right, if it exists,
        // or the one to the left, it exists. If neither exist, then it is just
        // the same as sheet one index, as there is only one sheet
        const sheetTwoIndex = sheetOneIndex + 1 <= props.sheetJSONArray.length - 1
            ? sheetOneIndex + 1
            : (sheetOneIndex - 1 >= 0 ? sheetOneIndex - 1 : sheetOneIndex);
        if (props.sheetJSONArray.length === 0) {
            // If there is no data, we just set default values
            this.state = {
                mergeType: MergeType.LOOKUP,
                sheetOneIndex: sheetOneIndex,
                sheetTwoIndex: sheetTwoIndex,
                sheetOneMergeKey: '',
                sheetTwoMergeKey: '',
                sheetOneSelectedColumns: [],
                sheetTwoSelectedColumns: [],
                sheetOneToggleAll: true,
                sheetTwoToggleAll: true,
                originalDfNames: props.dfNames,
                stepID: ''
            };
        }
        else {
            const suggestedKeys = getSuggestedColumnHeaderKeys(props.sheetJSONArray, sheetOneIndex, sheetTwoIndex);
            // We default to selecting _all_ columns
            const sheetOneSelectedColumns = [...this.props.sheetJSONArray[sheetOneIndex].columns];
            const sheetTwoSelectedColumns = [...this.props.sheetJSONArray[sheetTwoIndex].columns];
            this.state = {
                mergeType: MergeType.LOOKUP,
                sheetOneIndex: sheetOneIndex,
                sheetTwoIndex: sheetTwoIndex,
                sheetOneMergeKey: suggestedKeys.sheetOneMergeKey,
                sheetTwoMergeKey: suggestedKeys.sheetTwoMergeKey,
                sheetOneSelectedColumns: sheetOneSelectedColumns,
                sheetTwoSelectedColumns: sheetTwoSelectedColumns,
                sheetOneToggleAll: true,
                sheetTwoToggleAll: true,
                originalDfNames: props.dfNames,
                stepID: ''
            };
        }
        this.setNewSheetIndex = this.setNewSheetIndex.bind(this);
        this.setNewMergeKey = this.setNewMergeKey.bind(this);
        this.toggleKeepColumnHeader = this.toggleKeepColumnHeader.bind(this);
        this.toggleAll = this.toggleAll.bind(this);
        this.sendMergeMessage = this.sendMergeMessage.bind(this);
        this.setNewMergeType = this.setNewMergeType.bind(this);
    }
    /*
        It is best practice to make async calls not in the constructor,
        but rather in the componentDidMount lifecycle method!
    */
    componentDidMount() {
        // Send the first merge message there are at least 2 sheets in Mito.
        if (this.props.sheetJSONArray.length >= 2) {
            void this.sendMergeMessage();
        }
    }
    /*
        Helper function for updating the merge type in state and
        sending the merge message to the backend.
    */
    setNewMergeType(newMergeType) {
        this.setState({ mergeType: newMergeType }, () => {
            void this.sendMergeMessage();
        });
    }
    /*
        When one of the two merge indexes is changed, we change state by:
        1. Updating the sheet index.
        2. Updating the columns that are selected (this defaults to all), as well as the toggle
           for this sheet.
        3. Trying to find a new merge key between the sheets
    */
    setNewSheetIndex(sheetNumber, newSheetIndex) {
        const indexName = sheetNumber == MergeSheet.First ? 'sheetOneIndex' : 'sheetTwoIndex';
        const selectedColumnsName = sheetNumber == MergeSheet.First ? 'sheetOneSelectedColumns' : 'sheetTwoSelectedColumns';
        const toggleAllName = sheetNumber == MergeSheet.First ? 'sheetOneToggleAll' : 'sheetTwoToggleAll';
        const newSelectedColumns = [...this.props.sheetJSONArray[newSheetIndex].columns];
        this.setState(prevState => {
            // Return if we're not changing anything!
            if (prevState[indexName] == newSheetIndex) {
                return;
            }
            const newSuggestedKeys = getSuggestedColumnHeaderKeys(this.props.sheetJSONArray, sheetNumber === MergeSheet.First ? newSheetIndex : prevState.sheetOneIndex, sheetNumber === MergeSheet.Second ? newSheetIndex : prevState.sheetTwoIndex);
            return Object.assign(Object.assign({}, prevState), { [indexName]: newSheetIndex, [selectedColumnsName]: newSelectedColumns, [toggleAllName]: true, sheetOneMergeKey: newSuggestedKeys.sheetOneMergeKey, sheetTwoMergeKey: newSuggestedKeys.sheetTwoMergeKey });
        }, () => {
            void this.sendMergeMessage();
        });
    }
    /*
        Sets a new merge key for one of the merge sheets
    */
    setNewMergeKey(sheetNumber, newMergeKey) {
        const mergeKeyName = sheetNumber === MergeSheet.First ? 'sheetOneMergeKey' : 'sheetTwoMergeKey';
        this.setState(prevState => {
            return Object.assign(Object.assign({}, prevState), { [mergeKeyName]: newMergeKey });
        }, () => {
            void this.sendMergeMessage();
        });
    }
    /*
        Toggles if we should keep a specific column at from one of the sheets in the new
        merged sheet.
    */
    toggleKeepColumnHeader(sheetNumber, columnHeader) {
        const selectedColumnsName = sheetNumber == MergeSheet.First ? 'sheetOneSelectedColumns' : 'sheetTwoSelectedColumns';
        const mergeKeyName = sheetNumber == MergeSheet.First ? 'sheetOneMergeKey' : 'sheetTwoMergeKey';
        this.setState(prevState => {
            // We the don't let you toggle the merge key!
            if (prevState[mergeKeyName] === columnHeader) {
                return;
            }
            const newSelectedColumns = [...prevState[selectedColumnsName]];
            if (newSelectedColumns.includes(columnHeader)) {
                newSelectedColumns.splice(newSelectedColumns.indexOf(columnHeader), 1);
            }
            else {
                newSelectedColumns.push(columnHeader);
            }
            return Object.assign(Object.assign({}, prevState), { [selectedColumnsName]: newSelectedColumns });
        }, () => {
            void this.sendMergeMessage();
        });
    }
    /*
        Toggles all the check boxes selected for a given sheet, based on the current
        toggleAll state for that sheet.
    */
    toggleAll(sheetNumber) {
        const sheetIndexName = sheetNumber === MergeSheet.First ? 'sheetOneIndex' : 'sheetTwoIndex';
        const toggleAllName = sheetNumber === MergeSheet.First ? 'sheetOneToggleAll' : 'sheetTwoToggleAll';
        const selectedColumnsName = sheetNumber === MergeSheet.First ? 'sheetOneSelectedColumns' : 'sheetTwoSelectedColumns';
        this.setState(prevState => {
            let newSelectedColumns = [];
            if (!prevState[toggleAllName]) {
                newSelectedColumns = [...this.props.sheetJSONArray[this.state[sheetIndexName]].columns];
            }
            return Object.assign(Object.assign({}, prevState), { [selectedColumnsName]: newSelectedColumns, [toggleAllName]: !prevState[toggleAllName] });
        }, () => {
            void this.sendMergeMessage();
        });
    }
    /*
        Completes the merge operation by sending information for the merge
        to the backend, potentially overwriting what's already there!
    */
    async sendMergeMessage() {
        // NOTE: We make sure to send the merge keys in the selected columns, no matter what
        const sheetOneSelectedColumns = [...this.state.sheetOneSelectedColumns];
        if (!sheetOneSelectedColumns.includes(this.state.sheetOneMergeKey)) {
            sheetOneSelectedColumns.push(this.state.sheetOneMergeKey);
        }
        const sheetTwoSelectedColumns = [...this.state.sheetTwoSelectedColumns];
        if (!sheetTwoSelectedColumns.includes(this.state.sheetTwoMergeKey)) {
            sheetTwoSelectedColumns.push(this.state.sheetTwoMergeKey);
        }
        const stepID = await this.props.mitoAPI.sendMergeMessage(this.state.mergeType, this.state.sheetOneIndex, this.state.sheetOneMergeKey, sheetOneSelectedColumns, this.state.sheetTwoIndex, this.state.sheetTwoMergeKey, sheetTwoSelectedColumns, this.state.stepID);
        // Save the merge ID
        this.setState({ stepID: stepID });
    }
    render() {
        /*
            If there are less than 2 sheets in Mito to merge together,
            then display this error message.
        */
        if (this.props.sheetJSONArray.length < 2) {
            return (React.createElement(DefaultTaskpane, { header: 'Merge Sheets Together', setCurrOpenTaskpane: this.props.setCurrOpenTaskpane, taskpaneBody: React.createElement(Fragment, null, "Make sure you have at least two dataframes in Mito before merging.") }));
        }
        /*
            We don't let you select or unselect the sheet merge key, and note that we must account
            for the shift in the indexes that this causes when updating if the state of an item is
            toggled.

            Thus, we filter out the merge keys from both the list of columns, as well as the
            toggles for these columns.
        */
        const sheetOneOriginalColumns = this.props.sheetJSONArray[this.state.sheetOneIndex].columns;
        const sheetTwoOriginalColumns = this.props.sheetJSONArray[this.state.sheetTwoIndex].columns;
        const sheetOneColumns = sheetOneOriginalColumns.filter(columnHeader => columnHeader !== this.state.sheetOneMergeKey);
        const sheetTwoColumns = sheetTwoOriginalColumns.filter(columnHeader => columnHeader !== this.state.sheetTwoMergeKey);
        const sheetOneToggles = sheetOneColumns.map(columnHeader => this.state.sheetOneSelectedColumns.includes(columnHeader));
        const sheetTwoToggles = sheetTwoColumns.map(columnHeader => this.state.sheetTwoSelectedColumns.includes(columnHeader));
        return (React.createElement(DefaultTaskpane, { header: 'Merge Sheets Together', setCurrOpenTaskpane: this.props.setCurrOpenTaskpane, taskpaneBody: React.createElement(Fragment, null,
                React.createElement(Row, { justify: 'space-between', align: 'center', suppressTopBottomMargin: true },
                    React.createElement(Col, { offsetRight: 1 },
                        React.createElement("p", { className: 'text-header-3' }, "Merge Type")),
                    React.createElement(Col, null,
                        React.createElement(Select, { value: this.state.mergeType, options: Object.values(MergeType), subtexts: {
                                [MergeType.LOOKUP]: "Includes all rows from the first sheet and only matching rows from the second sheet. If there are mulitple matches in the second sheet, only takes the first.",
                                [MergeType.LEFT]: "Includes all rows from the first sheet and only matching rows from the second sheet. Includes all matches.",
                                [MergeType.RIGHT]: "Includes all rows from the second sheet and only matching rows from the  first sheet. Includes all matches.",
                                [MergeType.INNER]: "Only includes rows that have matches in both sheets.",
                                [MergeType.OUTER]: "Includes all rows from both sheets, regardless of whether there is a match in the other sheet.",
                            }, onChange: (mergeType) => {
                                const newMergeTypeEnum = mergeType;
                                this.setNewMergeType(newMergeTypeEnum);
                            }, width: 'medium' }))),
                React.createElement(MergeSheetAndKeySelection, { sheetNum: MergeSheet.First, dfNames: this.state.originalDfNames, sheetIndex: this.state.sheetOneIndex, otherSheetIndex: this.state.sheetTwoIndex, setNewSheetIndex: (newSheetIndex) => { this.setNewSheetIndex(MergeSheet.First, newSheetIndex); }, columnHeaders: this.props.sheetJSONArray[this.state.sheetOneIndex].columns, mergeKey: this.state.sheetOneMergeKey, setNewMergeKey: (newMergeKey) => this.setNewMergeKey(MergeSheet.First, newMergeKey) }),
                React.createElement("p", { className: 'text-header-3' }, "Columns to Keep"),
                React.createElement(MultiToggleBox, { itemsArray: sheetOneColumns, itemToggleState: sheetOneToggles, toggleItemAtIndex: (index) => this.toggleKeepColumnHeader(MergeSheet.First, sheetOneColumns[index]), toggleAllOptions: {
                        currentToggleAllState: this.state.sheetOneToggleAll,
                        toggleAll: () => { this.toggleAll(MergeSheet.First); }
                    } }),
                React.createElement(MergeSheetAndKeySelection, { sheetNum: MergeSheet.Second, dfNames: this.state.originalDfNames, sheetIndex: this.state.sheetTwoIndex, otherSheetIndex: this.state.sheetOneIndex, setNewSheetIndex: (newSheetIndex) => { this.setNewSheetIndex(MergeSheet.Second, newSheetIndex); }, columnHeaders: this.props.sheetJSONArray[this.state.sheetTwoIndex].columns, mergeKey: this.state.sheetTwoMergeKey, setNewMergeKey: (newMergeKey) => this.setNewMergeKey(MergeSheet.Second, newMergeKey) }),
                React.createElement("div", null,
                    React.createElement("p", { className: 'text-header-3' }, "Columns to Keep"),
                    React.createElement(MultiToggleBox, { itemsArray: sheetTwoColumns, itemToggleState: sheetTwoToggles, toggleItemAtIndex: (index) => this.toggleKeepColumnHeader(MergeSheet.Second, sheetTwoColumns[index]), toggleAllOptions: {
                            currentToggleAllState: this.state.sheetTwoToggleAll,
                            toggleAll: () => { this.toggleAll(MergeSheet.Second); }
                        } }))) }));
    }
}
export default MergeTaskpane;
//# sourceMappingURL=MergeTaskpane.js.map