// Copyright (c) Mito
import React, { Fragment, useEffect, useRef, useState } from 'react';
import DefaultTaskpane from '../DefaultTaskpane';
import PivotTableKeySelection from './PivotTableKeySelection';
// Import 
import PivotTableValueSelection from './PivotTableValueSelection';
import usePrevious from '../../../hooks/usePrevious';
import Select from '../../elements/Select';
import Row from '../../spacing/Row';
import Col from '../../spacing/Col';
import { allDfNamesToSelectableDfNameToSheetIndex, getDeduplicatedArray, valuesArrayToRecord, valuesRecordToArray } from './pivotUtils';
// NOTE: these aggregation functions need to be supported
// in mitosheet/steps/pivot.py as well
export var AggregationType;
(function (AggregationType) {
    AggregationType["SUM"] = "sum";
    AggregationType["MEAN"] = "mean";
    AggregationType["MEDIAN"] = "median";
    AggregationType["STD"] = "std";
    AggregationType["MIN"] = "min";
    AggregationType["MAX"] = "max";
    AggregationType["COUNT"] = "count";
    AggregationType["COUNT_UNIQUE"] = "count unique";
})(AggregationType || (AggregationType = {}));
const PivotTaskpane = (props) => {
    const [stepID, setStepID] = useState(undefined);
    // We save the dataframe names upon creation of the pivot table
    // so that the user cannot switch to the pivot table they are editing
    const [selectableDfNameToSheetIndex] = useState(allDfNamesToSelectableDfNameToSheetIndex(props.dfNames, props.destinationSheetIndex));
    /*
        If the pivot table is being opened with existing editing params, then
        we set all the starting values to these parameters. Note that some of them
        may be invalid, but we display errors in the respective sections of the pivot
        table if this is the case.

        Furthermore, we take special care with the sheet index to make sure that it
        is not out of bounds. So, if it is, we default it to the final sheet. This
        might lead to a wacky pivot, but that's better than crashing the sheet.

        NOTE: we use a single object for the entire state of the pivot table so that
        we can update the entire object at once, which is useful when we're refreshing
        the params to the pivot table on an update.

        ALSO NOTE: the type of the pivotValues is different between the backend and
        the frontend, due to it being easier to manipulate as an array on the frontend.
        We use the utils `valuesRecordToArray` and `valuesArraytoRecord` to convert
        between these two formats.
    */
    const [pivotState, setPivotState] = useState({
        selectedSheetIndex: props.existingPivotParams === undefined
            ? props.selectedSheetIndex
            : Math.min(props.existingPivotParams.sheet_index, props.sheetJSONArray.length - 1),
        pivotRows: props.existingPivotParams === undefined ? [] : props.existingPivotParams.pivot_rows,
        pivotColumns: props.existingPivotParams === undefined ? [] : props.existingPivotParams.pivot_columns,
        pivotValues: props.existingPivotParams === undefined ? [] : valuesRecordToArray(props.existingPivotParams.values)
    });
    // Save the last step index, so that we can check if an undo occured
    const prevLastStepIndex = usePrevious(props.lastStepIndex);
    /*
        Completes the pivot operation by sending information for the pivoting
        to the backend, making so to respect the stepId
    */
    const sendPivotTableUpdateMessage = async () => {
        const _stepID = await props.mitoAPI.sendPivotMessage(pivotState.selectedSheetIndex, 
        // Deduplicate the rows and columns before sending them to the backend
        // as otherwise this generates errors if you have duplicated key
        getDeduplicatedArray(pivotState.pivotRows), getDeduplicatedArray(pivotState.pivotColumns), 
        // Convert pivotValues back to the state that the backend expects
        valuesArrayToRecord(pivotState.pivotValues), props.destinationSheetIndex, stepID);
        setStepID(_stepID);
    };
    /*
        A callback used by the select data source Select Element so that it can
        set the state of the Pivot Table Taskpane
    */
    const setSelectedSheet = (newSheetName) => {
        const newSelectedSheetIndex = selectableDfNameToSheetIndex[newSheetName];
        // If you didn't select a new sheet, then don't do clear your selections
        if (newSelectedSheetIndex == pivotState.selectedSheetIndex) {
            return;
        }
        // Set the selected index, and clear the pivot table
        setPivotState({
            selectedSheetIndex: newSelectedSheetIndex,
            pivotRows: [],
            pivotColumns: [],
            pivotValues: []
        });
    };
    const addPivotValueAggregation = (columnHeader) => {
        const newPivotValues = [...pivotState.pivotValues];
        newPivotValues.push([columnHeader, AggregationType.COUNT]);
        setPivotState({
            selectedSheetIndex: pivotState.selectedSheetIndex,
            pivotRows: pivotState.pivotRows,
            pivotColumns: pivotState.pivotColumns,
            pivotValues: newPivotValues
        });
    };
    const removePivotValueAggregation = (valueIndex) => {
        const newPivotValues = [...pivotState.pivotValues];
        newPivotValues.splice(valueIndex, 1);
        setPivotState({
            selectedSheetIndex: pivotState.selectedSheetIndex,
            pivotRows: pivotState.pivotRows,
            pivotColumns: pivotState.pivotColumns,
            pivotValues: newPivotValues
        });
    };
    const editPivotValueAggregation = (valueIndex, newAggregationType, newColumnHeader) => {
        const newPivotValues = [...pivotState.pivotValues];
        newPivotValues[valueIndex] = [newColumnHeader, newAggregationType];
        setPivotState({
            selectedSheetIndex: pivotState.selectedSheetIndex,
            pivotRows: pivotState.pivotRows,
            pivotColumns: pivotState.pivotColumns,
            pivotValues: newPivotValues
        });
    };
    const addKey = (rowOrColumn, columnHeader) => {
        let newColumnHeaders = [];
        if (rowOrColumn === 'row') {
            newColumnHeaders = [...pivotState.pivotRows];
        }
        else {
            newColumnHeaders = [...pivotState.pivotColumns];
        }
        newColumnHeaders.push(columnHeader);
        if (rowOrColumn === 'row') {
            setPivotState({
                selectedSheetIndex: pivotState.selectedSheetIndex,
                pivotRows: newColumnHeaders,
                pivotColumns: pivotState.pivotColumns,
                pivotValues: pivotState.pivotValues
            });
        }
        else {
            setPivotState({
                selectedSheetIndex: pivotState.selectedSheetIndex,
                pivotRows: pivotState.pivotRows,
                pivotColumns: newColumnHeaders,
                pivotValues: pivotState.pivotValues
            });
        }
    };
    const removeKey = (rowOrColumn, keyIndex) => {
        let newColumnHeaders = [];
        if (rowOrColumn === 'row') {
            newColumnHeaders = [...pivotState.pivotRows];
        }
        else {
            newColumnHeaders = [...pivotState.pivotColumns];
        }
        newColumnHeaders.splice(keyIndex, 1);
        if (rowOrColumn === 'row') {
            setPivotState({
                selectedSheetIndex: pivotState.selectedSheetIndex,
                pivotRows: newColumnHeaders,
                pivotColumns: pivotState.pivotColumns,
                pivotValues: pivotState.pivotValues
            });
        }
        else {
            setPivotState({
                selectedSheetIndex: pivotState.selectedSheetIndex,
                pivotRows: pivotState.pivotRows,
                pivotColumns: newColumnHeaders,
                pivotValues: pivotState.pivotValues
            });
        }
    };
    const editKey = (rowOrColumn, keyIndex, newColumnHeader) => {
        let newColumnHeaders = [];
        if (rowOrColumn === 'row') {
            newColumnHeaders = [...pivotState.pivotRows];
        }
        else {
            newColumnHeaders = [...pivotState.pivotColumns];
        }
        newColumnHeaders[keyIndex] = newColumnHeader;
        if (rowOrColumn === 'row') {
            setPivotState({
                selectedSheetIndex: pivotState.selectedSheetIndex,
                pivotRows: newColumnHeaders,
                pivotColumns: pivotState.pivotColumns,
                pivotValues: pivotState.pivotValues
            });
        }
        else {
            setPivotState({
                selectedSheetIndex: pivotState.selectedSheetIndex,
                pivotRows: pivotState.pivotRows,
                pivotColumns: newColumnHeaders,
                pivotValues: pivotState.pivotValues
            });
        }
    };
    /*
        Taken from https://stackoverflow.com/questions/53253940/make-react-useeffect-hook-not-run-on-initial-render
        
        We make sure that if you are editing a pivot table, the first time
        that useEffect runs, it does not send a message to update the pivot
        table.

        This means that opening a pivot table with edits on it does not
        automatically overwrite the edits that you made to that pivot
        table. Insted, you need to begin editing the pivot table for it
        then to delete these edits to the pivot.
    */
    const ignoreNextUpdate = useRef(props.existingPivotParams !== undefined);
    useEffect(() => {
        if (ignoreNextUpdate.current) {
            ignoreNextUpdate.current = false;
            return;
        }
        // We do not send a pivot message if there is no actual
        // data in the sheet to pivot
        if (props.sheetJSONArray.length > 0) {
            void sendPivotTableUpdateMessage();
        }
    }, [pivotState]);
    const refreshParamsAfterUndoOrRedo = async () => {
        // If there is a desintation index, then we are editing other that, otherwise
        // we are just editing the last sheet 
        const currentDestinationSheetIndex = props.destinationSheetIndex !== undefined
            ? props.destinationSheetIndex
            : props.sheetJSONArray.length - 1;
        const params = await props.mitoAPI.getPivotParams(currentDestinationSheetIndex);
        // If we get the params, set them to the params of this pivot table. 
        if (params !== undefined) {
            // We also set ignoreNextUpdate to true, so that these updates
            // don't cause a new message to get sent, as then we get trapped
            // in an infinite loop where an undo will cause a new pivot message
            // to get sent 
            ignoreNextUpdate.current = true;
            setPivotState({
                selectedSheetIndex: params.sheet_index,
                pivotRows: params.pivot_rows,
                pivotColumns: params.pivot_columns,
                pivotValues: valuesRecordToArray(params.values)
            });
        }
    };
    useEffect(() => {
        // If there has been an undo or redo, then we refresh the params to this pivot
        if (prevLastStepIndex && prevLastStepIndex !== props.lastStepIndex) {
            void refreshParamsAfterUndoOrRedo();
        }
    }, [props.lastStepIndex]);
    /*
        If there is no possible Pivot taskpane that can be displayed (e.g. the sheetJSON is empty),
        give an error message indicating so.
    */
    if (props.sheetJSONArray.length === 0) {
        return (React.createElement(DefaultTaskpane, { header: 'Create a Pivot Table', setCurrOpenTaskpane: props.setCurrOpenTaskpane, taskpaneBody: React.createElement(Fragment, null, "Please import data before pivoting.") }));
    }
    const columnHeaders = props.sheetJSONArray[pivotState.selectedSheetIndex].columns;
    return (React.createElement(DefaultTaskpane, { header: props.destinationSheetIndex ?
            `Edit Pivot Table ${props.dfNames[props.destinationSheetIndex]}`
            : `Create Pivot Table ${props.dfNames[props.dfNames.length - 1]}`, setCurrOpenTaskpane: props.setCurrOpenTaskpane, taskpaneBody: React.createElement(Fragment, null,
            React.createElement(Row, { justify: 'space-between', align: 'center' },
                React.createElement(Col, null,
                    React.createElement("p", { className: 'text-header-3' }, "Data Source")),
                React.createElement(Col, null,
                    React.createElement(Select, { width: 'medium', value: props.dfNames[pivotState.selectedSheetIndex], options: Object.keys(selectableDfNameToSheetIndex), onChange: setSelectedSheet }))),
            React.createElement("div", { className: 'default-taskpane-body-section-div' },
                React.createElement(PivotTableKeySelection, { sectionTitle: 'Rows', columnHeaders: columnHeaders, selectedColumnHeaders: pivotState.pivotRows, addKey: (columnHeader) => { addKey('row', columnHeader); }, removeKey: (keyIndex) => { removeKey('row', keyIndex); }, editKey: (keyIndex, newColumnHeader) => { editKey('row', keyIndex, newColumnHeader); }, mitoAPI: props.mitoAPI, rowOrColumn: 'row' })),
            React.createElement("div", { className: 'default-taskpane-body-section-div' },
                React.createElement(PivotTableKeySelection, { sectionTitle: 'Columns', columnHeaders: columnHeaders, selectedColumnHeaders: pivotState.pivotColumns, addKey: (columnHeader) => { addKey('column', columnHeader); }, removeKey: (keyIndex) => { removeKey('column', keyIndex); }, editKey: (keyIndex, newColumnHeader) => { editKey('column', keyIndex, newColumnHeader); }, mitoAPI: props.mitoAPI, rowOrColumn: 'column' })),
            React.createElement("div", { className: 'default-taskpane-body-section-div' },
                React.createElement(PivotTableValueSelection, { columnHeaders: columnHeaders, pivotValues: pivotState.pivotValues, addPivotValueAggregation: addPivotValueAggregation, removePivotValueAggregation: removePivotValueAggregation, editPivotValueAggregation: editPivotValueAggregation, mitoAPI: props.mitoAPI }))) }));
};
export default PivotTaskpane;
//# sourceMappingURL=PivotTaskpane.js.map